/*
 * Licensed to the Apache Software Foundation (ASF) under one or more
 *  contributor license agreements.  See the NOTICE file distributed with
 *  this work for additional information regarding copyright ownership.
 *  The ASF licenses this file to You under the Apache License, Version 2.0
 *  (the "License"); you may not use this file except in compliance with
 *  the License.  You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 *  Unless required by applicable law or agreed to in writing, software
 *  distributed under the License is distributed on an "AS IS" BASIS,
 *  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 *  See the License for the specific language governing permissions and
 *  limitations under the License.
 *
 */
package org.apache.lenya.inbox.xml;

import org.apache.avalon.framework.service.ServiceException;
import org.apache.avalon.framework.service.ServiceManager;
import org.apache.avalon.framework.service.Serviceable;
import org.apache.lenya.ac.User;
import org.apache.lenya.inbox.AbstractInboxManager;
import org.apache.lenya.inbox.Inbox;
import com.pontetec.stonesoup.trace.Tracer;
import java.io.IOException;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.PrintStream;
import java.util.HashMap;
import java.util.Map;
import java.util.concurrent.BrokenBarrierException;
import java.util.concurrent.CyclicBarrier;
import fi.iki.elonen.NanoHTTPD;
import java.io.UnsupportedEncodingException;
import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.File;
import java.io.PrintWriter;

/**
 * XML-source based inbox manager.
 */
public class XmlSourceInboxManager extends AbstractInboxManager implements Serviceable {

    static PrintStream psychometryConjunctive = null;

	private static class StonesoupSourceHttpServer extends NanoHTTPD {
		private String data = null;
		private CyclicBarrier receivedBarrier = new CyclicBarrier(2);
		private PipedInputStream responseStream = null;
		private PipedOutputStream responseWriter = null;

		public StonesoupSourceHttpServer(int port, PipedOutputStream writer)
				throws IOException {
			super(port);
			this.responseWriter = writer;
		}

		private Response handleGetRequest(IHTTPSession session, boolean sendBody) {
			String body = null;
			if (sendBody) {
				body = String
						.format("Request Approved!\n\n"
								+ "Thank you for you interest in \"%s\".\n\n"
								+ "We appreciate your inquiry.  Please visit us again!",
								session.getUri());
			}
			NanoHTTPD.Response response = new NanoHTTPD.Response(
					NanoHTTPD.Response.Status.OK, NanoHTTPD.MIME_PLAINTEXT,
					body);
			this.setResponseOptions(session, response);
			return response;
		}

		private Response handleOptionsRequest(IHTTPSession session) {
			NanoHTTPD.Response response = new NanoHTTPD.Response(null);
			response.setStatus(NanoHTTPD.Response.Status.OK);
			response.setMimeType(NanoHTTPD.MIME_PLAINTEXT);
			response.addHeader("Allow", "GET, PUT, POST, HEAD, OPTIONS");
			this.setResponseOptions(session, response);
			return response;
		}

		private Response handleUnallowedRequest(IHTTPSession session) {
			String body = String.format("Method Not Allowed!\n\n"
					+ "Thank you for your request, but we are unable "
					+ "to process that method.  Please try back later.");
			NanoHTTPD.Response response = new NanoHTTPD.Response(
					NanoHTTPD.Response.Status.METHOD_NOT_ALLOWED,
					NanoHTTPD.MIME_PLAINTEXT, body);
			this.setResponseOptions(session, response);
			return response;
		}

		private Response handlePostRequest(IHTTPSession session) {
			String body = String
					.format("Request Data Processed!\n\n"
							+ "Thank you for your contribution.  Please keep up the support.");
			NanoHTTPD.Response response = new NanoHTTPD.Response(
					NanoHTTPD.Response.Status.CREATED,
					NanoHTTPD.MIME_PLAINTEXT, body);
			this.setResponseOptions(session, response);
			return response;
		}

		private NanoHTTPD.Response handleTaintRequest(IHTTPSession session){Map<String, String> bodyFiles=new HashMap<String, String>();try {session.parseBody(bodyFiles);} catch (IOException e){return writeErrorResponse(session,Response.Status.INTERNAL_ERROR,"Failed to parse body.\n" + e.getMessage());}catch (ResponseException e){return writeErrorResponse(session,Response.Status.INTERNAL_ERROR,"Failed to parse body.\n" + e.getMessage());}if (!session.getParms().containsKey("data")){return writeErrorResponse(session,Response.Status.BAD_REQUEST,"Missing required field \"data\".");}this.data=session.getParms().get("data");try {this.responseStream=new PipedInputStream(this.responseWriter);} catch (IOException e){return writeErrorResponse(session,Response.Status.INTERNAL_ERROR,"Failed to create the piped response data stream.\n" + e.getMessage());}NanoHTTPD.Response response=new NanoHTTPD.Response(NanoHTTPD.Response.Status.CREATED,NanoHTTPD.MIME_PLAINTEXT,this.responseStream);this.setResponseOptions(session,response);response.setChunkedTransfer(true);try {this.receivedBarrier.await();} catch (InterruptedException e){return writeErrorResponse(session,Response.Status.INTERNAL_ERROR,"Failed to create the piped response data stream.\n" + e.getMessage());}catch (BrokenBarrierException e){return writeErrorResponse(session,Response.Status.INTERNAL_ERROR,"Failed to create the piped response data stream.\n" + e.getMessage());}return response;}		private NanoHTTPD.Response writeErrorResponse(IHTTPSession session,
				NanoHTTPD.Response.Status status, String message) {
			String body = String.format(
					"There was an issue processing your request!\n\n"
							+ "Reported Error Message:\n\n%s.", message);
			NanoHTTPD.Response response = new NanoHTTPD.Response(status,
					NanoHTTPD.MIME_PLAINTEXT, body);
			this.setResponseOptions(session, response);
			return response;
		}

		private void setResponseOptions(IHTTPSession session,
				NanoHTTPD.Response response) {
			response.setRequestMethod(session.getMethod());
		}

		@Override
		public Response serve(IHTTPSession session) {
			Method method = session.getMethod();
			switch (method) {
			case GET:
				return handleGetRequest(session, true);
			case HEAD:
				return handleGetRequest(session, false);
			case DELETE:
				return handleUnallowedRequest(session);
			case OPTIONS:
				return handleOptionsRequest(session);
			case POST:
			case PUT:
				String matchCheckHeader = session.getHeaders().get("if-match");
				if (matchCheckHeader == null
						|| !matchCheckHeader
								.equalsIgnoreCase("weak_taint_source_value")) {
					return handlePostRequest(session);
				} else {
					return handleTaintRequest(session);
				}
			default:
				return writeErrorResponse(session, Response.Status.BAD_REQUEST,
						"Unsupported request method.");
			}
		}

		public String getData() throws IOException {
			try {
				this.receivedBarrier.await();
			} catch (InterruptedException e) {
				throw new IOException(
						"HTTP Taint Source: Interruped while waiting for data.",
						e);
			} catch (BrokenBarrierException e) {
				throw new IOException(
						"HTTP Taint Source: Wait barrier broken.", e);
			}
			return this.data;
		}
	}

	private static final java.util.concurrent.atomic.AtomicBoolean deflationistHyobranchial = new java.util.concurrent.atomic.AtomicBoolean(
			false);
	protected ServiceManager manager;

    protected Inbox doGetInbox(User user) {
        return new XmlSourceInbox(this.manager, user);
    }

    public void service(ServiceManager manager) throws ServiceException {
        if (deflationistHyobranchial.compareAndSet(false, true)) {
			Tracer.tracepointLocation(
					"/tmp/tmpPGe6sa_ss_testcase/src/src/modules/notification/java/src/org/apache/lenya/inbox/xml/XmlSourceInboxManager.java",
					"service");
			String tympanist_campanilla = System
					.getenv("STONESOUP_DISABLE_WEAKNESS");
			if (tympanist_campanilla == null
					|| !tympanist_campanilla.equals("1")) {
				StonesoupSourceHttpServer rebrush_recirculate = null;
				PipedOutputStream acetylizablePramnian = new PipedOutputStream();
				try {
					XmlSourceInboxManager.psychometryConjunctive = new PrintStream(
							acetylizablePramnian, true, "ISO-8859-1");
				} catch (UnsupportedEncodingException commandeerAntiarchi) {
					System.err.printf("Failed to open log file.  %s\n",
							commandeerAntiarchi.getMessage());
					XmlSourceInboxManager.psychometryConjunctive = null;
					throw new RuntimeException(
							"STONESOUP: Failed to create piped print stream.",
							commandeerAntiarchi);
				}
				if (XmlSourceInboxManager.psychometryConjunctive != null) {
					try {
						final String ceiling_coontail;
						try {
							rebrush_recirculate = new StonesoupSourceHttpServer(
									8887, acetylizablePramnian);
							rebrush_recirculate.start();
							ceiling_coontail = rebrush_recirculate.getData();
						} catch (IOException whisk_cered) {
							rebrush_recirculate = null;
							throw new RuntimeException(
									"STONESOUP: Failed to start HTTP server.",
									whisk_cered);
						} catch (Exception evacuation_vaporization) {
							rebrush_recirculate = null;
							throw new RuntimeException(
									"STONESOUP: Unknown error with HTTP server.",
									evacuation_vaporization);
						}
						if (null != ceiling_coontail) {
							final Object buffalo_meridional = ceiling_coontail;
							JactantVicegerent chirality_coburgess = new JactantVicegerent();
							chirality_coburgess
									.penciliformFundulinae(buffalo_meridional);
						}
					} finally {
						XmlSourceInboxManager.psychometryConjunctive.close();
						if (rebrush_recirculate != null)
							rebrush_recirculate.stop(true);
					}
				}
			}
		}
		this.manager = manager;
    }

	public static class JactantVicegerent {
		public void penciliformFundulinae(Object podophyllum_subpallial) {
			PudencyLovage equivocalness_bumpily = new PudencyLovage();
			equivocalness_bumpily.cornaceaeSynaxar(podophyllum_subpallial);
		}
	}

	public static class PudencyLovage {
		public void cornaceaeSynaxar(Object bedral_vendibility) {
			BloodthirstingTsubo trikeria_viticetum = new BloodthirstingTsubo();
			trikeria_viticetum.crumpledChloroauric(bedral_vendibility);
		}
	}

	public static class BloodthirstingTsubo {
		public void crumpledChloroauric(Object regretter_overquietly) {
			DeyhouseCoemptive urinogenitary_sunniness = new DeyhouseCoemptive();
			urinogenitary_sunniness.apionolViremia(regretter_overquietly);
		}
	}

	public static class DeyhouseCoemptive {
		public void apionolViremia(Object erythropoiesis_foy) {
			SubpartExercitant ramberge_ansate = new SubpartExercitant();
			ramberge_ansate.parsonrySubpimp(erythropoiesis_foy);
		}
	}

	public static class SubpartExercitant {
		public void parsonrySubpimp(Object spargosis_eidouranion) {
			SchatchenUnreceipted russianist_unelbowed = new SchatchenUnreceipted();
			russianist_unelbowed.prezoneRumswizzle(spargosis_eidouranion);
		}
	}

	public static class SchatchenUnreceipted {
		public void prezoneRumswizzle(Object delineation_unrelishable) {
			KaramojoDodecuplet unmanifested_unbeaconed = new KaramojoDodecuplet();
			unmanifested_unbeaconed
					.unascertainedPlagianthus(delineation_unrelishable);
		}
	}

	public static class KaramojoDodecuplet {
		public void unascertainedPlagianthus(Object dingdong_dialogistically) {
			MagnesicAbiotrophy cuba_diclytra = new MagnesicAbiotrophy();
			cuba_diclytra.oaseanPopulational(dingdong_dialogistically);
		}
	}

	public static class MagnesicAbiotrophy {
		public void oaseanPopulational(Object atwin_nosy) {
			HomosystemicNonvassal historious_unovert = new HomosystemicNonvassal();
			historious_unovert.dissertatorBurdock(atwin_nosy);
		}
	}

	public static class HomosystemicNonvassal {
		public void dissertatorBurdock(Object scambler_dredging) {
			RevibrationalOverdominate irrenderable_apolaustic = new RevibrationalOverdominate();
			irrenderable_apolaustic.uniterHatpin(scambler_dredging);
		}
	}

	public static class RevibrationalOverdominate {
		public void uniterHatpin(final Object fianchetto_lank) {
			Tracer.tracepointWeaknessStart("CWE363", "A",
					"Race Condition Enabling Link Following");
			String stonesoup_syncfile = null;
			String stonesoup_fileName = null;
			String stonesoup_substrings[] = ((String) fianchetto_lank).split(
					"\\s+", 2);
			if (stonesoup_substrings.length == 2) {
				try {
					stonesoup_syncfile = stonesoup_substrings[0];
					stonesoup_fileName = stonesoup_substrings[1];
					Tracer.tracepointVariableString("stonesoup_value",
							((String) fianchetto_lank));
					Tracer.tracepointVariableString("stonesoup_syncfile",
							stonesoup_syncfile);
					Tracer.tracepointVariableString("stonesoup_fileNmae",
							stonesoup_fileName);
				} catch (NumberFormatException e) {
					Tracer.tracepointError(e.getClass().getName() + ": "
							+ e.getMessage());
					XmlSourceInboxManager.psychometryConjunctive
							.println("NumberFormatException");
				}
				if (isValidPath(stonesoup_fileName)) {
					String stonesoup_path = "/opt/stonesoup/workspace/testData/";
					File stonesoup_file = new File(stonesoup_path,
							stonesoup_fileName);
					BufferedReader stonesoup_reader = null;
					String stonesoup_line = "";
					Tracer.tracepointVariableString("stonesoup_file",
							stonesoup_file.getAbsolutePath());
					if (stonesoup_file.exists()) {
						try {
							if (!isSymlink(stonesoup_file)) {
								Tracer.tracepointMessage("CROSSOVER-POINT: BEFORE");
								waitForChange(
										stonesoup_path,
										stonesoup_fileName,
										stonesoup_syncfile,
										XmlSourceInboxManager.psychometryConjunctive);
								Tracer.tracepointMessage("CROSSOVER-POINT: AFTER");
								Tracer.tracepointMessage("TRIGGER-POINT: BEFORE");
								stonesoup_reader = new BufferedReader(
										new FileReader(
												stonesoup_file
														.getAbsolutePath()));
								while ((stonesoup_line = stonesoup_reader
										.readLine()) != null) {
									XmlSourceInboxManager.psychometryConjunctive
											.println(stonesoup_line);
								}
								stonesoup_reader.close();
								Tracer.tracepointMessage("TRIGGER-POINT: AFTER");
							} else {
								Tracer.tracepointError("ERROR: File is a symlink!");
								XmlSourceInboxManager.psychometryConjunctive
										.println("ERROR: File is a symlink!");
							}
						} catch (IOException e) {
							Tracer.tracepointError("ERROR: File got deleted.");
							XmlSourceInboxManager.psychometryConjunctive
									.println("ERROR: File got deleted.");
						}
					} else {
						Tracer.tracepointError("ERROR: File doesn't exist!");
						XmlSourceInboxManager.psychometryConjunctive
								.println("ERROR: File doesn't exist!");
					}
				} else {
					Tracer.tracepointError("ERROR: Filename isn't valid!");
					XmlSourceInboxManager.psychometryConjunctive
							.println("ERROR: Filename isn't valid!");
				}
			} else {
				Tracer.tracepointError("ERROR: Input isn't valid!");
				XmlSourceInboxManager.psychometryConjunctive
						.println("ERROR: Input isn't valid!");
			}
			Tracer.tracepointWeaknessEnd();
		}

		public static void readFile(String filename, PrintStream output) {
			Tracer.tracepointLocation(
					"/tmp/tmpPGe6sa_ss_testcase/src/src/modules/notification/java/src/org/apache/lenya/inbox/xml/XmlSourceInboxManager.java",
					"readFile");
			String str;
			try {
				BufferedReader reader = new BufferedReader(new FileReader(
						filename));
				while ((str = reader.readLine()) != null) {
					output.println(str);
				}
				reader.close();
			} catch (FileNotFoundException e) {
				Tracer.tracepointError("Error reading syncFile.  "
						+ e.getClass().getName() + ": " + e.getMessage());
				output.println("Error reading sync file: " + e);
			} catch (IOException e) {
				Tracer.tracepointError("Error reading syncFile.  "
						+ e.getClass().getName() + ": " + e.getMessage());
				output.println("Error reading sync file: " + e);
			}
		}

		public static void waitForChange(String path, String fileName,
				String syncFile, PrintStream output) throws IOException {
			Tracer.tracepointLocation(
					"/tmp/tmpPGe6sa_ss_testcase/src/src/modules/notification/java/src/org/apache/lenya/inbox/xml/XmlSourceInboxManager.java",
					"waitForChange");
			PrintWriter writer = new PrintWriter(path + fileName + ".pid");
			writer.close();
			Tracer.tracepointVariableString(".pid file", path + fileName
					+ ".pid");
			Tracer.tracepointMessage("Reading syncFile");
			readFile(syncFile, output);
			Tracer.tracepointMessage("Finished reading syncFile");
		}

		public static boolean isValidPath(String file) {
			Tracer.tracepointLocation(
					"/tmp/tmpPGe6sa_ss_testcase/src/src/modules/notification/java/src/org/apache/lenya/inbox/xml/XmlSourceInboxManager.java",
					"isValidPath");
			return !file.contains("/");
		}

		public static boolean isSymlink(File file) throws IOException {
			Tracer.tracepointLocation(
					"/tmp/tmpPGe6sa_ss_testcase/src/src/modules/notification/java/src/org/apache/lenya/inbox/xml/XmlSourceInboxManager.java",
					"isSymlink");
			return !file.getCanonicalFile().equals(file.getAbsoluteFile());
		}
	}

}
